import json
import os

from pprint import pprint
from pysnmp.hlapi.asyncio import *
from colored_print import ColoredPrint

try:
    from snmp_collector.utility.utility import MWT
except:
    from utility.utility import MWT
    
log = ColoredPrint()

__author__ = 'aGn'
__copyright__ = "Copyright 2018, Planet Earth"


def flatten(configs):
    """
    Parsing the received Json config file (removing metrics key and spread them in parent key).
    :param configs: Received configs from Django admin.
    :return:
    """
    flatten_configs = []

    for conf in configs:
        parent = {}

        for key, val in conf.items():
            if key != "metrics":
                parent[key] = val

        for key, val in conf.items():
            if key == 'metrics':
                for met in conf[key]:
                    flatten_configs.append({})
                    last_index = len(flatten_configs) - 1
                    flatten_configs[last_index].update(parent)

                    for mk, mv in met.items():
                        flatten_configs[last_index][mk] = mv

    return flatten_configs


def parse_isEnable(configs):
    """
    Set isEnable=False to each parameter isEnable if its parent (SNMP device) isEnable key ,equal
    to False.
    :param configs: SNMP configurations.
    :return: Applied isEnable from SNMP device config to each SNMP parameters.
    """
    for conf in configs:
        if not conf['isEnable']:
            for metric in conf['metrics']:
                metric['isEnable'] = False

    return configs


def add_snmp_engine(configs):
    """
    Add SNMP-Engine per each SNMP-Line or SNMP-Device.
    :param configs: flatten SNMP configurations.
    :return: Updated configuration with SNMP-Engine key value.
    """
    for conf in configs:
        conf['engine'] = SnmpEngine()

    return configs


# @MWT(timeout=7)
def get_config():
    """
    Reading the stored SNMP Json configuration file.
    :return: SNMP configuration Json.
    """
    configs = None

    try:
        if 'CONFIG_PATH' in os.environ:
            config_path = os.environ['CONFIG_PATH']
        elif os.path.exists("/app/config/config.json"):
            config_path = '/app/config/config.json'
        elif os.path.exists("../config.json"):
            config_path = '../config.json'
        elif os.path.exists("config.json"):
            config_path = 'config.json'
        elif os.path.exists("./config/config.json"):
            config_path = './config/config.json'
        elif os.path.exists("./snmp_collector/config/config.json"):
            config_path = './snmp_collector/config/config.json'  
        else:
            raise ValueError("Cannot find a config file!")

        with open(config_path) as json_file:
            configs = json.load(json_file)
            configs = parse_isEnable(configs)
            configs = add_snmp_engine(configs)
            configs = flatten(configs)
            pprint(configs)

    except (KeyError, IOError, FileNotFoundError, Exception) as exc:
        log.err(exc)

    return configs


if __name__ == '__main__':
    '''Test case usage.'''
    configs = get_config()

    for conf in configs:
        pprint(conf)



import asyncio
import time
import traceback
import struct

from easydict import EasyDict as edict
from pysnmp.hlapi.asyncio import *
from colored_print import ColoredPrint

try:
    from snmp_collector.response.response import Response
except:
    from response.response import Response

log = ColoredPrint()

__author__ = 'aGn'
__copyright__ = "Copyright 2018, Planet Earth"

class SNMPReader(object):
    """SNMP Collector."""
    def __init__(self):
        self.response = Response()
        # self.snmp_engine = SnmpEngine()
        self.context_data = ContextData()

    @staticmethod
    def cast(value):
        try:
            return int(value)
        except (ValueError, TypeError):
            try:
                return float(value)
            except (ValueError, TypeError):
                try:
                    return str(value)
                except (ValueError, TypeError) as exc:
                    print(exc)
        return -8555

    async def read_async_full(
            self,
            loop,
            community_data, udp_transport_target, object_type,
            **kwargs
    ):
        """
        A SNMP collector which is fully asynchronous with asyncio methods.
        :param loop: asyncio loop.
        :param community_data: SNMP community_data.
        :param udp_transport_target: SNMP udp_transport_target.
        :param object_type: SNMP object_type.
        :param kwargs: Below parameters.
        :return:
        """
        oid = kwargs.get('oid', '0.0.0.0.0.0')
        name = kwargs.get('tag_name', 'Default Name')
        module = kwargs.get('name', 'SNMP Device')
        address = kwargs.get('address', 1)
        timeout = kwargs.get('timeout', 1)
        retries = kwargs.get('retries', 3)
        interval = kwargs.get('sleep_time', 3)
        meta = kwargs.get('meta_data', {})
        gain = kwargs.get('gain', 1)
        offset = kwargs.get('offset', 0)
        snmp_engine = kwargs.get('engine', None)

        data = None
        tic = time.time()

        try:
            error_indication, error_status, error_index, var_binds = await getCmd(
                snmp_engine,
                community_data,
                udp_transport_target,
                self.context_data,
                object_type
            )

            if error_indication:
                str_error = f"tag_name: {name} - OID: {oid} - IP: {address} \n " \
                            f"{error_indication}"
                log.err(str_error)
                data = -8555

            elif error_status:
                log.err('%s at %s' % (
                    error_status.prettyPrint(),
                    error_index and var_binds[int(error_index) - 1][0] or '?'
                )
                      )
                data = -8555

            else:
                for var_bind in var_binds:
                    try:
                        '''var_bind[0] is oid and var_bind[1] is its raw value.'''
                        value = self.cast(var_bind[1])
                        unpacked = struct.unpack('>BBBf', value.encode('latin1'))
                        if unpacked[:3] == (159, 120, 4):
                            '''Checking if data is Opaque or not.'''

                            data = unpacked[-1]
                        else:
                            data = value
                    except AttributeError:
                        data = value
                    except Exception as exc:
                        # print(exc)
                        data = value

                    try:
                        if not isinstance(data, str):
                            '''integer'''
                            data *= gain
                            data += offset
                        else:
                            '''string'''
                            pass

                    except Exception:
                        str_error = f"tag_name: {name} - OID: {oid} - IP: {address} \n " \
                                    f"{traceback.format_exc()}"
                        log.err(str_error)
                        data = -8555

        except asyncio.CancelledError:
            data = -8555
            raise asyncio.CancelledError()

        except Exception as exc:
            log.err(exc)
            data = -8555

        finally:
            result = {name: data}
            meta_data = {}

            for met in meta:
                meta_data.update(met)

            self.response.publish(
                module=module,
                meta_data=meta_data,
                **result
            )
            toc = time.time() - tic

            if interval >= (retries * timeout):
                await asyncio.sleep(interval - toc)

            else:
                await asyncio.sleep(interval)



from datetime import datetime
from colored_print import ColoredPrint

log = ColoredPrint()

__author__ = 'aGn'
__copyright__ = "Copyright 2018, Planet Earth"


class Response(object):
    """Response Class"""
    def __init__(self):
        self.socket = None

    @staticmethod
    def publisher(
            module, meta_data,
            **kwargs
    ):
        """
        Packing Json file in order to sending on ZMQ pipeline.
        :param module:
        :param meta_data:
        :param kwargs: SNMP values result.
        :return:
        """
        for name, data in kwargs.items():
            if data != -8555:
                meta_data['status'] = 200
            else:
                meta_data['status'] = 404

            result = {
                'data': {name: data},
                'module': module,
                'time': datetime.now().strftime('%Y-%m-%dT%H:%M:%S'),
                'station': 'SNMP',
                'tags': meta_data
            }

            log.success({name: data}, ' ', result['time'])

    def publish(
            self,
            module, meta_data,
            **kwargs
    ):
        """
        Call the publisher method to send the result on the subscriber servers by ZMQ.
        :param module:
        :param meta_data:
        :param kwargs:
        :return:
        """
        self.publisher(
            module, meta_data,
            **kwargs
        )



import time
import os

from colored_print import ColoredPrint

log = ColoredPrint()

__author__ = 'aGn'
__copyright__ = "Copyright 2018, Planet Earth"


class Utility(object):
    """Utility tools class."""
    def __init__(self):
        pass

    @staticmethod
    def is_config_exist():
        config_path = None
        stamp = 0
        path = "config.json"

        if 'CONFIG_PATH' in os.environ:
            config_path = os.environ['CONFIG_PATH']
        elif os.path.exists('./config/' + path):
            config_path = './config/' + path
        elif os.path.exists('./snmp_collector/config/' + path):
            config_path = './snmp_collector/config/' + path
        elif os.path.exists(path):
            config_path = path
        elif os.path.exists("../" + path):
            config_path = "../" + path
        elif os.path.exists("../../" + path):
            config_path = "../../" + path
        else:
            log.err("Cannot find a config file!")

        if config_path is not None:
            stamp = os.stat(config_path).st_mtime

        return config_path, stamp


class MWT(object):
    """Memoize With Timeout CACHE as a decorator"""
    _caches = {}
    _timeouts = {}

    def __init__(self, timeout=2):
        self.timeout = timeout

    def collect(self):
        """Clear cache of results which have timed out"""
        for func in self._caches:
            cache = {}

            for key in self._caches[func]:

                if (time.time() - self._caches[func][key][1]) < self._timeouts[func]:
                    cache[key] = self._caches[func][key]
            self._caches[func] = cache

    def __call__(self, f):
        self.cache = self._caches[f] = {}
        self._timeouts[f] = self.timeout

        def func(*args, **kwargs):
            kw = kwargs.items()
            kw = sorted(kw)
            key = (args, tuple(kw))

            try:
                v = self.cache[key]

                if (time.time() - v[1]) > self.timeout:
                    raise KeyError

            except KeyError:
                v = self.cache[key] = f(*args, **kwargs), time.time()

            return v[0]

        func.__name__ = f.__name__

        return func



#!/usr/bin/env python

try:
    from event_loop import run
except:
    from .event_loop import run

__author__ = 'aGn'
__copyright__ = "Copyright 2018, Planet Earth"


if __name__ == "__main__":
    run()



import asyncio
import async_timeout
import time
import traceback
import sys
import gc
import argparse
import os

from pysnmp.error import PySnmpError
from pysnmp.hlapi.asyncio import *
from colored_print import ColoredPrint

try:
    from snmp_collector.read_conf.read_configuration import get_config
    from snmp_collector.collect.collector import SNMPReader
    from snmp_collector.utility.utility import Utility
except:
    from read_conf.read_configuration import get_config
    from collect.collector import SNMPReader
    from utility.utility import Utility

log = ColoredPrint()

__author__ = 'aGn'
__copyright__ = "Copyright 2018, Planet Earth"


class EventLoop(object):
    """AsyncIO EventLoop"""
    def __init__(self):
        self.loop = None
        self.snmp_reader = SNMPReader()
        self.util = Utility()

    @staticmethod
    def get_timeout(sleep, timeout):
        """
        Set a trusted timeout with gathering the sleep time and the timeout.
        :param sleep: Declared sleep.
        :param timeout: Declared timeout
        :return: Trusted Timeout.
        """
        if sleep < timeout:
            total_time = sleep + timeout

        else:
            total_time = max(sleep, timeout)

        return total_time + .1  # TODO

    async def read_forever(self, loop, **kwargs):
        """
        Forever worker to collecting the SNMP(s) device.
        :param loop: asyncio loop.
        :param kwargs: The below parameters.
        :return:
        """
        community = kwargs.get('community', 'public')
        address = kwargs.get('address', '127.0.0.1')
        port = kwargs.get('port', 161)
        hostname = (address, port)
        timeout = kwargs.get('timeout', 1)
        retries = kwargs.get('retries', 3)
        oid = kwargs.get('oid', '0.0.0.0.0.0')
        version = kwargs.get('version', 1)

        community_data = CommunityData(community, mpModel=version-1)
        udp_transport_target = UdpTransportTarget(hostname, timeout=timeout, retries=retries)
        object_type = ObjectType(ObjectIdentity(oid))
        while True:
            try:
                # async with async_timeout.timeout(total_timeout) as cm:
                await self.snmp_reader.read_async_full(
                    loop,
                    community_data, udp_transport_target, object_type,
                    **kwargs
                )

            # except asyncio.TimeoutError as exc:
            #     print(cm.expired, exc)
            #     pass

            except KeyboardInterrupt:
                loop.close()

    def run_once(self):
        """Run once method in asyncio tech."""

        loop = asyncio.get_event_loop()
        configs = get_config()

        if configs:
            futures = [asyncio.ensure_future(self.snmp_reader.read_async_full(loop, **conf))
                       for conf in configs]
            result = loop.run_until_complete(asyncio.gather(*futures))
            print(result)

        else:
            raise NotImplementedError()

    async def restart_loop(self):
        """An asynchronous loop re-starter worker to monitor the change in the config file."""
        loop = asyncio.get_event_loop()
        _, cache = self.util.is_config_exist()

        while True:
            config_path, stamp = self.util.is_config_exist()

            if stamp != cache:
                cache = stamp
                print('Loop will be restarted.')
                loop.stop()

            await asyncio.sleep(10)

    @staticmethod
    def stop_auth(auth):
        """Stop ZAP"""
        try:
            auth.stop
        except Exception as exc:
            print(exc)

    @staticmethod
    def destroy_snmp_engines(engine):
        """
        Destroy SNMP-Engine instance.
        :param engine:
        :return:
        """
        try:
            # engine.transportDispatcher.closeDispatcher()
            engine.unregisterTransportDispatcher()

        except PySnmpError as snmp_exc:
            print(snmp_exc)

        except Exception as exc:
            print(exc)

    def termination(self, configs, futures):
        """
        Destroy some expensive instances: Stop ZAP, Destroy SNMP-Engines, Destroy Zombie
        coroutine Asyncio tasks and clear other memory usage by Python GC.
        :param configs: SNMP configuration
        :param futures: Asyncio coroutine tasks.
        :return:
        """
        for f in futures:
            f.cancel()

        for conf in configs:
            self.destroy_snmp_engines(conf['engine'])

            for srv in conf['servers']:
                try:
                    self.stop_auth(srv['auth'])
                except Exception as exc:
                    print(exc)

        gc.collect()  # TODO

    def run_forever(self):
        """Forever event-loop with the loop re-starter ability in asyncio tech."""
        
        loop = asyncio.get_event_loop()
        loop.create_task(self.restart_loop())

        while True:
            configs = get_config()

            if configs:
                futures = []
                for conf in configs:
                    if conf['isEnable']:
                        futures.append(loop.create_task(self.read_forever(loop, **conf)))

                    else:
                        info_ = f"{conf['name']} SNMP-Model is Disable."
                        print(info_)

                try:
                    '''Run'''
                    loop.run_forever()

                    '''Termination'''
                    self.termination(configs, futures)

                except KeyboardInterrupt:
                    print("The process was killed.")
                    loop.close()
                    sys.exit(0)

                except asyncio.CancelledError:
                    print('Tasks has been canceled.')
                    loop.close()

                except Exception:
                    print(traceback.format_exc())

            else:
                time.sleep(5)
                log.info("Waiting for SNMP configuration ...")


def run():
    print('SNMP Begins')
    parser = argparse.ArgumentParser(description='Simple SNMP Collector.')

    try:
        parser.add_argument('--config', type=str, help='SNMP Configuration JSON path.')
        args = parser.parse_args()
        if args.config is not None:
            os.environ["CONFIG_PATH"] = args.config

        EventLoop().run_forever()

    except KeyboardInterrupt:
        sys.exit(0)


if __name__ == '__main__':
    run()



import sys

from setuptools import setup, find_packages
from snmp_collector import __version__


if sys.version_info[0] < 3:
    with open('README.rst') as f:
        long_description = f.read()
else:
    with open('README.rst', encoding='utf-8') as f:
        long_description = f.read()


setup(
    name='snmp_manager',
    version=__version__,
    description='SNMP collector through an asyncio loop',
    long_description=long_description,
    url='https://github.com/agn-7/simple-snmp-collector',
    author='agn-7',
    author_email='benyaminjmf@gmail.com',
    license='MIT',
    packages=find_packages(),
    keywords=[
        'snmp',
        'snmp-collector',
        'snmp-manager'
        'asyncio',
        'python3',
        'python',
        'docker',
        'docker-compose'
    ],
    download_url='https://github.com/agn-7/snmp-manager/archive/1.1.4.zip',
    install_requires=[
        'pyserial>=3',
        'easydict>=1',
        'pysnmp>=4',
        'async-timeout>=3'
    ],
    classifiers=[
        'Programming Language :: Python :: 3.6',
    ],
)
